"""Fitting function to evaluate with the fitting algorithm."""
import functools
from dataclasses import InitVar, dataclass, field
from typing import Callable, Dict, Optional

import numpy as np

from eddington.exceptions import FitFunctionRuntimeError
from eddington.fit_functions_registry import FitFunctionsRegistry


@dataclass(unsafe_hash=True)
class FitFunction:  # pylint: disable=invalid-name,too-many-instance-attributes
    """
    Fitting function class.

    This class wraps up a callable which gets 2 parameters:

    * ``a`` - An array with the parameters of the function.
    * ``x`` - The sample data to be fit.

    Our main goal is to find the best suitable ``a`` that match given ``x`` values to
    given ``y`` values.

    :param fit_func: The actual fitting function.
    :type fit_func: callable
    :param n: Number of parameters. the length of "a" in `fit_func`.
    :type n: int
    :param name: The name of the function.
    :type name: str
    :param syntax: The syntax of the fitting function
    :type syntax: str
    :param a_derivative: a function representing the derivative of fit_func according
     to the "a" array
    :type a_derivative: callable
    :param x_derivative: a function representing the derivative of fit_func according
     to x
    :type a_derivative: callable
    :param title_name: same as `name` but in title case
    :type title_name: str
    :param save: Should this function be saved in the :class:`FitFunctionsRegistry`
    :type save: bool
    """

    fit_func: Callable = field(repr=False)
    n: int = field(repr=False)
    name: Optional[str] = field()
    syntax: Optional[str] = field(default=None)
    a_derivative: Optional[Callable] = field(default=None, repr=False)
    x_derivative: Optional[Callable] = field(default=None, repr=False)
    title_name: str = field(init=False, repr=False)
    fixed: Dict[int, float] = field(init=False, repr=False)
    save: InitVar[bool] = True

    def __post_init__(self, save):
        """Post init methods."""
        self.title_name = self.__get_title_name()
        self.fixed = dict()
        self.x_derivative = self.__wrap_x_derivative(self.x_derivative)
        self.a_derivative = self.__wrap_a_derivative(self.a_derivative)
        if save:
            FitFunctionsRegistry.add(self)

    def __get_title_name(self):
        return self.name.title().replace("_", " ")

    def __validate_parameters_number(self, a):
        a_length = len(a)
        if a_length != self.n:
            raise FitFunctionRuntimeError(
                f"Input length should be {self.active_parameters}, "
                f"got {a_length - len(self.fixed)}"
            )

    def __call__(self, *args):
        """Call the fit function as a regular callable."""
        a, x = self.__extract_a_and_x(args)
        self.__validate_parameters_number(a)
        return self.fit_func(a, x)

    def assign(self, a):
        """Assign the function parameters."""
        a = self.__add_fixed_values(a)
        self.__validate_parameters_number(a)
        self.fixed = dict(enumerate(a))
        return self

    def fix(self, index, value):
        """
        Fix parameter with predefined value.

        :param index: The index of the parameter to fix. Starting from 0
        :type index: int
        :param value: The value to fix
        :type value: float
        :return: self :class:`FitFunction`
        """
        if index < 0 or index >= self.n:
            raise FitFunctionRuntimeError(
                f"Cannot fix index {index}. "
                f"Indices should be between 0 and {self.n - 1}"
            )
        self.fixed[index] = value
        return self

    def unfix(self, index):
        """
        Unfix a fixed parameter.

        :param index: The index of the parameter to unfix
        :type index: int
        :return: self :class:`FitFunction`
        """
        del self.fixed[index]
        return self

    def clear_fixed(self):
        """
        Clear all fixed parameters.

        :return: self :class:`FitFunction`
        """
        self.fixed.clear()

    @property
    def signature(self):
        """Same as name."""
        return self.name

    @property
    def active_parameters(self):
        """Number of active parameters (aka, unfixed)."""
        return self.n - len(self.fixed)

    def __wrap_x_derivative(self, method):
        if method is None:
            return None

        @functools.wraps(method)
        def wrapper(*args):
            a, x = self.__extract_a_and_x(args)
            self.__validate_parameters_number(a)
            return method(a, x)

        return wrapper

    def __wrap_a_derivative(self, method):
        if method is None:
            return None

        @functools.wraps(method)
        def wrapper(*args):
            a, x = self.__extract_a_and_x(args)
            self.__validate_parameters_number(a)
            result = method(a, x)
            if len(self.fixed) == 0:
                return result
            return np.delete(result, list(self.fixed.keys()), axis=0)

        return wrapper

    def __extract_a_and_x(self, args):
        if len(args) == 0:
            raise FitFunctionRuntimeError(
                f'No parameters has been given to "{self.name}"'
            )
        if len(args) == 1:
            a = [self.fixed[i] for i in sorted(self.fixed.keys())]
            x = args[0]
        else:
            a = self.__add_fixed_values(args[0])
            x = args[1]
        return a, x

    def __add_fixed_values(self, a):
        for i in sorted(self.fixed.keys()):
            a = np.insert(a, i, self.fixed[i])
        return a


def fit_function(  # pylint: disable=invalid-name,too-many-arguments
    n, name=None, syntax=None, a_derivative=None, x_derivative=None, save=True
):
    """
    Wrapper making a simple callable into a :class:`FitFunction`.

    :param n: Number of parameters. The length of parameter ``a`` of the fitting
     function.
    :type n: int
    :param name: The name of the function.
    :type name: str
    :param syntax: The syntax of the fitting function.
    :type syntax: str
    :param a_derivative: a function representing the derivative of the fit function
     according to the "a" parameter array
    :type a_derivative: callable
    :param x_derivative: a function representing the derivative of the fit function
     according to x
    :param save: Should this function be saved in the
     :class:`FitFunctionsRegistry`
    :type save: bool
    :return: :class:`FitFunction` instance.
    """

    def wrapper(func):
        func_name = func.__name__ if name is None else name
        if func.__doc__ is None:
            func.__doc__ = ""
        func.__doc__ += f"""

Syntax: :code:`y = {syntax}`

:param a: Coefficients array of length {n}
:type a: ``numpy.ndarray``
:param x: Free parameter
:type x: ``numpy.ndarray`` or ``float``
:returns: ``numpy.ndarray`` or ``float``
    """
        return functools.wraps(func)(
            FitFunction(
                fit_func=func,
                n=n,
                name=func_name,
                syntax=syntax,
                a_derivative=a_derivative,
                x_derivative=x_derivative,
                save=save,
            )
        )

    return wrapper



# pylint: disable=missing-class-docstring
"""Exception classes for Eddington."""


class EddingtonException(Exception):  # noqa: D101
    pass


# Fit Function Errors


class FitFunctionLoadError(EddingtonException):  # noqa: D101
    pass


class FitFunctionSaveError(EddingtonException):  # noqa: D101
    pass


class FitFunctionRuntimeError(EddingtonException):  # noqa: D101
    pass


# Fit Data Errors


class FitDataError(EddingtonException):  # noqa: D101
    pass


class FitDataInvalidFile(FitDataError):  # noqa: D101
    pass


class FitDataInvalidFileSyntax(FitDataInvalidFile):  # noqa: D101
    def __init__(self, file_name, sheet=None):  # noqa: D107
        sheet_msg = "" if sheet is None else f' in sheet "{sheet}"'
        msg = f'"{file_name}" has invalid syntax{sheet_msg}.'
        super().__init__(msg)


class FitDataColumnsLengthError(FitDataInvalidFile):  # noqa: D101
    msg = "All columns in FitData should have the same length"

    def __init__(self):  # noqa: D107
        super().__init__(self.msg)


class FitDataColumnIndexError(FitDataError):  # noqa: D101
    def __init__(self, index, max_index):  # noqa: D107
        super().__init__(
            f"No column number {index} in data. "
            f"index should be between 1 and {max_index}"
        )


class FitDataColumnExistenceError(FitDataError):  # noqa: D101
    def __init__(self, column):  # noqa: D107
        super().__init__(f'Could not find column "{column}" in data')


class FitDataColumnsSelectionError(FitDataError):  # noqa: D101
    pass



"""Core functionalities of the Eddington platform."""
from eddington.exceptions import (
    EddingtonException,
    FitDataColumnExistenceError,
    FitDataColumnIndexError,
    FitDataColumnsLengthError,
    FitDataError,
    FitDataInvalidFile,
    FitDataInvalidFileSyntax,
    FitFunctionLoadError,
    FitFunctionRuntimeError,
)
from eddington.fit_data import FitData
from eddington.fit_function_class import FitFunction, fit_function
from eddington.fit_functions_list import (
    constant,
    cos,
    exponential,
    hyperbolic,
    inverse_power,
    linear,
    normal,
    parabolic,
    poisson,
    polynomial,
    sin,
    straight_power,
)
from eddington.fit_functions_registry import FitFunctionsRegistry
from eddington.fit_result import FitResult
from eddington.fitting import fit_to_data

__all__ = [
    # Fit functions infrastructure
    "FitFunction",
    "fit_function",
    "FitFunctionsRegistry",
    # Fit functions
    "constant",
    "exponential",
    "hyperbolic",
    "linear",
    "parabolic",
    "polynomial",
    "cos",
    "sin",
    "straight_power",
    "inverse_power",
    "normal",
    "poisson",
    # Fitting algorithm
    "fit_to_data",
    # Exceptions
    "EddingtonException",
    "FitFunctionRuntimeError",
    "FitFunctionLoadError",
    "FitDataError",
    "FitDataColumnExistenceError",
    "FitDataColumnIndexError",
    "FitDataInvalidFile",
    "FitDataColumnsLengthError",
    "FitDataInvalidFileSyntax",
    # Data structures
    "FitData",
    "FitResult",
]



"""Registry containing predefined fit functions and fit functions generators."""
from eddington.exceptions import FitFunctionLoadError, FitFunctionSaveError


class FitFunctionsRegistry:  # noqa: D415,D213,D205
    """A singleton class containing all saved :class:`FitFunction` instances."""

    __name_to_func_dict = dict()  # type: ignore

    @classmethod
    def add(cls, func):
        """Add a fit function."""
        if func.name in cls.__name_to_func_dict:
            raise FitFunctionSaveError(
                f'Cannot save "{func.name}" to registry'
                " since there is another fit function with this name"
            )
        cls.__name_to_func_dict[func.name] = func

    @classmethod
    def remove(cls, func_name):
        """Remove a fit function."""
        del cls.__name_to_func_dict[func_name]

    @classmethod
    def clear(cls):
        """Clear all fit functions and generators."""
        cls.__name_to_func_dict.clear()

    @classmethod
    def all(cls):
        """Get all fit functions and generators."""
        return cls.__name_to_func_dict.values()

    @classmethod
    def names(cls):
        """Names of all fit functions and generators."""
        return cls.__name_to_func_dict.keys()

    @classmethod
    def load(cls, name):
        """Get a fit function and generators by name."""
        if not cls.exists(name):
            raise FitFunctionLoadError(f"No fit function named {name}")
        return cls.__name_to_func_dict[name]

    @classmethod
    def exists(cls, func_name):
        """Checks whether a fit function exist."""
        return func_name in cls.__name_to_func_dict



"""Implementation of the fitting algorithm."""
import numpy as np
from scipy.odr import ODR, Model, RealData

from eddington import FitData, FitFunction, FitResult


def fit_to_data(
    data: FitData,
    func: FitFunction,
    a0: np.ndarray = None,
    use_x_derivative=True,
    use_a_derivative=True,
):  # pylint: disable=C0103
    """
    Implementation of the fitting algorithm using scipy ODR algorithm.

    :param data: :class:`FitData`. Fitting data to optimize
    :param func: :class:`FitFunction`. a function to fit the data according to.
    :param a0: nd.array. initail guess for the parameters
    :param use_x_derivative: Boolean. indicates whether to use x derviative or not.
    :param use_a_derivative: Boolean. indicates whether to use a derviative or not.
    :return: :class:`FitResult`
    """
    model = Model(
        **__get_odr_model_kwargs(
            func, use_x_derivative=use_x_derivative, use_a_derivative=use_a_derivative,
        )
    )
    a0 = __get_a0(n=func.active_parameters, a0=a0)
    real_data = RealData(x=data.x, y=data.y, sx=data.xerr, sy=data.yerr)
    odr = ODR(data=real_data, model=model, beta0=a0)
    output = odr.run()
    a = output.beta
    chi2 = output.sum_square  # pylint: disable=E1101
    degrees_of_freedom = len(data.x) - func.active_parameters
    return FitResult(
        a0=a0,
        a=a,
        aerr=output.sd_beta,
        acov=output.cov_beta,
        degrees_of_freedom=degrees_of_freedom,
        chi2=chi2,
    )


def __get_odr_model_kwargs(
    func, use_x_derivative=True, use_a_derivative=True,
):
    kwargs = dict(fcn=func)
    if use_a_derivative and func.a_derivative is not None:
        kwargs["fjacb"] = func.a_derivative
    if use_x_derivative and func.x_derivative is not None:
        kwargs["fjacd"] = func.x_derivative
    return kwargs


def __get_a0(n, a0=None):  # pylint: disable=invalid-name
    """
    Returns initial parameters for fitting algorithm.

    :param n: Number of parameters
    :param a0: Initial parameters value. Optional
    :return: nd.array
    """
    if a0 is not None:
        return a0
    return np.full(shape=n, fill_value=1.0)



#!/usr/bin/env python
import os

from setuptools import setup

READTHEDOCS = "READTHEDOCS"

version = "0.0.14"
on_rtd = os.environ.get(READTHEDOCS) == "True"
install_requires = []
if not on_rtd:
    install_requires = ["scipy >= 1.5.2", "numpy >= 1.19.1", "openpyxl >= 3.0.4"]
setup(version=version, install_requires=install_requires)
